#!/bin/sh
#
""":"
exec python $0 ${1+"$@"}
"""

import Rappture
from Rappture.tools import getCommandOutput as RapptureExec
from Rappture.tools import getDriverNumber  as getDriverNumber
import shutil
import os
import sys
import re
import getopt

def executeSubmit(runDir,
                  executable,
                  arguments,
                  venue,
                  nCores,
                  wallTime,
                  inputFiles,
                  manager,
                  streamOutput=True):
   runName = "pnprun"
   workDir = os.getcwd()
   if not runDir:
      runDir = workDir

   submitCommand = ["submit"]
   submitCommand.append("--runName")
   submitCommand.append(runName)
   if venue and venue != "":
      submitCommand.append("--venue")
      submitCommand.append(venue)
   if nCores and nCores > 1:
      submitCommand.append("--nCpus")
      submitCommand.append(str(nCores))
   if wallTime and wallTime != "":
      submitCommand.append("--wallTime")
      submitCommand.append(wallTime)
   if manager and manager != "":
      submitCommand.append("--manager")
      submitCommand.append(manager)

   if inputFiles:
      for inputFile in inputFiles:
         submitCommand.append("--inputfile")
         submitCommand.append(inputFile)

   submitCommand.append(executable)
   if arguments:
      for argument in arguments:
         submitCommand.append(str(argument))

   if runDir != workDir:
      os.chdir(runDir)

   exitStatus,submitStdout,submitStderr = RapptureExec(submitCommand,streamOutput)

   if runDir != workDir:
      os.chdir(workDir)

   return exitStatus,submitStdout,submitStderr


# function for creating run directory and deleting old run directory
def createDir(dirName):
    resultDirPath = os.path.join(os.getcwd(),dirName)
    if os.path.exists(resultDirPath):
        shutil.rmtree(resultDirPath)
    os.mkdir(resultDirPath)
    return resultDirPath

def printHelp(err=''):
    print err
    print """usage: python %s [-x executable] <driverFile>
    -x, --executable <executable_path>
    -h, --help
"""%(sys.argv[0])
    sys.exit()


def getMyOptions():
    longOpts = ["executable=","help"]
    shortOpts = "x:h"
    opts, args = getopt.getopt(sys.argv[1:], shortOpts, longOpts)

    driver = ''
    executable = ''

    re_xml = re.compile('(/?.*/)*driver[0-9]+.[Xx][Mm][Ll]')

    # place all of the extra arguments into the string input
    for arg in args:
        if re_xml.match(arg):
            driver = arg
        else:
            input = input + " " + arg

    # match options
    for o, v in opts:
        if o in ("-x", "--executable"):
            executable = v
        elif o in ("-h", "--help"):
            printHelp()

    if driver == '':
        error = "no driver file provided, use -h for help"
        # raise RuntimeError,error
        printHelp(error)

    return (executable,driver)

if __name__ == '__main__':

    print "simulation started"

    executable = ''
    driver = ''
    (executable,driver) = getMyOptions()

    # open XML file and find number of jobs to run
    lib = Rappture.library(driver)
    numprocs = lib.get('input.(inputs).(voltg).(max_steps).current')

    # job parameters
    nodes = int(numprocs)
    walltime = '4:00:00'
    jobName = 'pnp' + getDriverNumber(driver)

    # new directory for results
    resultDir = createDir(jobName)
    shutil.copy(driver,resultDir)

    executeSubmit(resultDir,
                  executable,
                  [driver],
                  'rcac-standby',
                  nodes,
                  walltime,
                  None,
                  None)

    re_runxml = re.compile("run\d+\.xml$",re.IGNORECASE)
    runFile = filter(re_runxml.search,os.listdir(resultDir))
    if runFile != []:
       idx = len(runFile)
       runFile.sort()
       runFilePath = os.path.join(resultDir,runFile[idx-1])
       shutil.copy(runFilePath,os.getcwd())
    else:
       error = "Could not find runfile in output, check %s" % (resultDir)
       raise RuntimeError,error

    # place std.out and std.err streams in a file
    re_stdout = re.compile(".*\.stdout$")
    outFiles = filter(re_stdout.search,os.listdir(resultDir))
    if outFiles != []:
       for outFile in outFiles:
          outFilePath = os.path.join(resultDir,outFile)
          os.system("cat %s" % (outFilePath))

    re_stderr = re.compile(".*\.stderr$")
    errFiles = filter(re_stderr.search,os.listdir(resultDir))
    if errFiles != []:
       for errFile in errFiles:
          errFilePath = os.path.join(resultDir,errFile)
          os.system("cat %s" % (errFilePath))

    if os.path.exists(resultDir):
       shutil.rmtree(resultDir)

    sys.exit()

