/***********************************************************************
09.07.2004
0) Reference : Gilson, Sharp, and Honig J. Compu. Chem. v9, p327 (1987).
1) Distribution of charges onto the grid 
   (Honig et al, J. Compu. Chem. v9, p327 (1987)) is NOT used.
2) Instead, Cloud-In-Cell (CIC) method is used.

09.15.2004
1) Completed assigning the charges into grid.

09.30.2004
1) Assigning the dielectric constants using simple method. I didn't use
Lee & Richards surface method.

10.05.2004
1) Chaged the charge of glutamiate ion into that of protonated glutamic acids
10.06.2004
1) make GRDLNG variable

11.08.2004
1) Found out that the Conc is not needed in PB equation, instead Conc is needed,
so Conc is changed into Conc.

11.17.2004
1) In Gridid, 0:bulk water, 1:protein, 2: ion exclusion region, 3: membrane
   4: channel water

11.19.2004
1) Add odd-even ordering algorithm along with half sweep updating

12.06.2004
1) Implement the charge antialiasing method. 
(Bruccoleri et al, J. Comp. Chem. v18, 268 (1997))

12.14.2004
1) debug: ndots[NDOTS_MAX] = {0.0,} ---> ndots[NDOTS_MAX] = {0,}
2) Debug: in Epsy[idx] ----> Epsy[idy]
3) Add Insight VDW radii parameters (default VDW radii in Delphi)
  
12.15.2004
1) Add AMBER94 partial charges for GLN, LEU, TRP.

12.22.2004
1) Set the zero flux boundary (reflecting boundary) to the side walls, 
	5: side wall
2) Use both the bulk and channel diffusion constant.

12.28.2004
1) Set up the lipid dipoles and assign them into grids.

01.04.2005
1) Rotate the side chains of arginines to place them in the interface 
between the water and the lipid bilayers.

01.06.2005
1) Set up randomly oriented dipoles in xy plane and tilted in z axis

02.24.2005
1) Change the boundary condition for the concentration.

03.03.2005
1) Correct the charge alias method for hydrogen.

07.08.2005
1) Found a bug: if(rr <= Chnrad2 && fabs(z) <= 0.5 * TUBLNG) -->
		if(rr <= Chnionrad2 && fabs(z) <= 0.5 * TUBLNG) Grdid[idp] = 4;

06.14.2006
1) Began integrating into Rappture

06.21.2006
1) Converted to C++ syntax
-changed fprint to cout
-changed calloc/free to new/delete
-changed my_math header
-removed my_matrix header (not necessary)

06.22.2006
1) major syntax changes complete
-calculations still work
-output files eliminated
-define to const 

06.27.2006
1) outputs viewable in GUI, dummy inputs only

06.28.2006
1) input commands for voltage work in GUI
2) fixed "current denisty" confusion with "current"

06.29.2006
1) unofficially done with input parameters

07.10.2006
1) encountered problems with C++ libraries and Rappture library,
   difficulty using <sstream> "stdlibc++.so.5 conflicts with stdlibc++.so.6"

07.14.2006
1) finished touching up GUI in xml file
2) added model graphic from Chemsketch

07.19.2006
1) moved to Nanohub workspace, sstream errors resolved

07.25.2006
1) finished concentration profiles in zy, xy, and 3D
2) waiting on changes to allow better representation of membrane and peptides

07.26.2006
1) decreased calculation volume by 25%, greatly increased speed of iterations
2) successfully ran on "very fine grid"
3) looked into junk value bug, occurred in fine grid while usually
   restricted to coarse grid

07.27.2006
1) fixed junk value error, did not initialize Nvolz array when changing from
calloc() to new syntax

08.01.2006
1) wrote in MPI calls, reworked inputs and outputs

08.04.2006
1) finished working pbs script, still problem with Rappture interface (will be fixed later)
2) addressing problems with MPI_Gather into the VERY larger master arrays

08.11.2006
1) figured out how to accomodate large arrays, using stacking of arrays
Ngrdx*Ngrdy x numprocs*Ngrdz
2) test program works, but not when integrated into calculation
*************************************************************************/

/* C++ libraries */
#include <iostream>
#include <math.h>
#include <time.h>
#include <string.h>
#include <sstream>
#include <mpi.h>

using namespace std;

/* Rappture Libraries */
#include "RpLibrary.h"

/* Set Flags */
typedef int flag;
const bool TRUE  =  1;
const bool FALSE =  0;
const bool DEBUG =  0;

flag Flag_pot_calc               = TRUE; 
flag Flag_cnc_calc               = TRUE;
flag Flag_set_side_wall_boundary = TRUE;
flag Flag_fixed_boundary         = TRUE;
flag Flag_periodic_boundary      = FALSE;
flag Calc_unstable               = FALSE;

/* system parameters */
const int NATOMS_MAX  =  1000;
const int NATOMS_RING =    16;
int Nrings; 
double Tublng;
double Tubrad_min;
double Tubrad_max;
double Tubcatchg;                 /*0.00714285714, 0.0142857143 */ 
double Tubanichg;                 /*0.00714285714, 0.0142857143 */ 
double TubDplLng;
double Tubatmrad;
double Tubeps;
double Memeps;

/* ion parameters */
double Topcnc;
double Botcnc;
double Catchg;
double Anichg;
double Catdfnblk;
double Anidfnblk;
double Catdfnchn;
double Anidfnchn;
double Ionrad;

/* grid parameters */
int Ngrdx;
int Ngrdy;
int Ngrdz;
double Grdlng;
double Gridrad_max2;

/* iteration parameters */
double Tol1;
double Tol2;
const double OMEGA_INI =     1.5;         /* Overrelaxation parameter */
const int DSPSTP       =    10;           /* printout per DSPSTP */
int myid;

/* water parameters */
const double WATEPS    =    80.0;         /* 78.5 */
double Chweps;

/* Definitions of constants */
const double COULCNST  =  4172.836038;    /* Kcal/mol,Fixed charge, 10^-10 (m)unit */
const double CNCNST    =     2.512940282; /* Ion concentration, Kcal/mol, 10^-10 (m) unit */
const double CRTCNST   =     9.64853419;  /* pA */
const double KAPPA0    =     2.91323146; 
const double EV2KCALPMOL =  23.0605502;
/*Definitions of physical constants */
const double KBT       =     0.59219;     /* 298K, Kcal/mol */

/* Charge antialias parameters */
const int NCAMNEIBORS =    20;
const int NSUBDIV     =     5;            /* a subdivision in each dimension */
const int NMINGRDS    =     8; 

/* global arrays and values */
double Atmx[NATOMS_MAX]={0.0,},Atmy[NATOMS_MAX]={0.0,},Atmz[NATOMS_MAX]={0.0,};
double Atmchg[NATOMS_MAX],Atmrad[NATOMS_MAX];

int Nbins,Natms_tub,Ngrdx_2,Ngrdy_2,Ngrdz_2;
long NgrdxNgrdy,NgrdyNgrdz,NgrdxNgrdyNgrdz;
int *Grdid,*Nvolz,*Nvolr;

double *Grdchg,*Epsx,*Epsy,*Epsz,*Phi,*Cdfn,*Adfn,*Ccnc,*Acnc,*Cflx,*Aflx;
double **Phi2d,**Ccnc2d,**Acnc2d;
double Grdvol,Watvol,Tub_watvol,Tot_ese,Tubrad_max2,Tubrad_min2;
double Tubionrad_max,Tubionrad_min,Tubionrad_max2,Tubionrad_min2,Tublng_hlf;
double Ikbt,Grdlng2,Igrdlng;

/****************************************************************\
*                                                                *
*           Set up external parameter and memory allocation      *	
*                                                                *
\****************************************************************/

void setup_params_memalloc() {
  
  /* Set up parameters */
  Ikbt = 1.0 / KBT;
  Igrdlng = 1.0 / Grdlng;
  Grdlng2 = (Grdlng * Grdlng);
  Grdvol = (Grdlng * Grdlng * Grdlng);
  Ngrdx_2 = (Ngrdx - 1) / 2;
  Ngrdy_2 = (Ngrdy - 1) / 2;
  Ngrdz_2 = (Ngrdz - 1) / 2;
  NgrdxNgrdy = (Ngrdx * Ngrdy);
  NgrdyNgrdz = (Ngrdy * Ngrdz); 
  NgrdxNgrdyNgrdz = (Ngrdx * Ngrdy * Ngrdz); 
  Gridrad_max2 = ((Ngrdx * Grdlng * Ngrdx * Grdlng) + (Ngrdy * Grdlng * Ngrdy * Grdlng)) / 4;
  
  /* Allocate memory */	
  Grdid = new int[NgrdxNgrdyNgrdz];	
  Grdchg = new double[NgrdxNgrdyNgrdz];	
  Epsx  = new double[(Ngrdx-1) * Ngrdy * Ngrdz];	
  Epsy  = new double[Ngrdx * (Ngrdy-1) * Ngrdz];	
  Epsz  = new double[Ngrdx * Ngrdy * (Ngrdz-1)];	
  Phi   = new double[NgrdxNgrdyNgrdz];	
  Cdfn  = new double[NgrdxNgrdyNgrdz];  
  Adfn  = new double[NgrdxNgrdyNgrdz];
  Ccnc  = new double[NgrdxNgrdyNgrdz];
  Acnc  = new double[NgrdxNgrdyNgrdz];
  Cflx  = new double[NgrdxNgrdyNgrdz];
  Aflx  = new double[NgrdxNgrdyNgrdz];
  Nvolz = new int[Ngrdz];
  Nvolr = new int[Ngrdx_2];

  Phi2d = new double*[NgrdxNgrdy];
  Ccnc2d = new double*[NgrdxNgrdy];
  Acnc2d = new double*[NgrdxNgrdy];
  for(int ij = 0; ij < NgrdxNgrdy; ij++) {
    Phi2d[ij] = new double[Ngrdz];
    Ccnc2d[ij] = new double[Ngrdz];
    Acnc2d[ij] = new double[Ngrdz];
  }

  for(int k = 0; k < Ngrdz; k++) {
    Nvolz[k] = 0;
  }
  for(int i = 0; i < Ngrdx_2; i++) {
    Nvolr[i] = 0;
  }
}

/****************************************************************\
*                                                                *
*                         Free Memory                            *	
*                                                                *
\****************************************************************/
void free_memory() {
  
  delete[] Grdid;
  delete[] Grdchg;	
  delete[] Epsx;
  delete[] Epsy;
  delete[] Epsz;
  delete[] Phi;
  delete[] Cdfn;  
  delete[] Adfn;
  delete[] Ccnc;  
  delete[] Acnc;
  delete[] Cflx; 
  delete[] Aflx; 
  delete[] Nvolz;
  delete[] Nvolr;
  
  for(int ij = 0; ij < NgrdxNgrdy; ij++) {
    delete[] Phi2d[ij];
    delete[] Ccnc2d[ij];
    delete[] Acnc2d[ij];
  }
  delete[] Phi2d;
  delete[] Ccnc2d;
  delete[] Acnc2d;
}

/**********************************************************\
*                                                          *
*     Read the coordinates of nanotube from minimized      *
*                    NAMD output file                      *	 
*                                                          *
\**********************************************************/

void read_pdb_file() {
  
  /* Channel parameters */
  Natms_tub      = NATOMS_RING * Nrings;	
  Tubrad_max2    = Tubrad_max * Tubrad_max;	
  Tubrad_min2    = Tubrad_min * Tubrad_min;
  Tubionrad_max  = Tubrad_max + Ionrad;	
  Tubionrad_min  = Tubrad_min - Ionrad;
  Tubionrad_max2 = Tubionrad_max * Tubionrad_max;	
  Tubionrad_min2 = Tubionrad_min * Tubionrad_min;	
  Tublng_hlf     = 0.5 * Tublng;	
}
	
/****************************************************************\
*                                                                *
*               Read Atomic charge and VDW radii                 *
*                                                                *
\****************************************************************/
void read_charges_vdwradii() {
  int m;
  double untlng = Tublng / Nrings;	
  double delang = 2.0 * M_PI / NATOMS_RING;    
  double sum;

  for(sum = 0.0, m = 0; m < Nrings; m++) for(int n = 0; n < NATOMS_RING; n += 2){ 
    int l = m * NATOMS_RING + n;	
    /* negative dipole end */	
    Atmx[l] = (Tubrad_min + Tubatmrad + 0.5) * cos(delang * n);
    Atmy[l] = (Tubrad_min + Tubatmrad + 0.5) * sin(delang * n);
    Atmz[l] =  (m + 0.5) * untlng - Tublng_hlf;
    Atmchg[l] = Tubanichg;
    Atmrad[l] = Tubatmrad;
    sum += Atmchg[l];
		
    /* positive dipole end */	
    Atmx[l+1] = (Tubrad_min + TubDplLng + Tubatmrad + 0.5) * cos(delang * n);
    Atmy[l+1] = (Tubrad_min + TubDplLng + Tubatmrad + 0.5) * sin(delang * n);
    Atmz[l+1] =  (m + 0.5) * untlng - Tublng_hlf;
    Atmchg[l+1] = Tubcatchg;
    Atmrad[l+1] = Tubatmrad;
    sum += Atmchg[l+1]; 
  }
  cout << "Total charge of cylinder = " << sum << endl;
}

/****************************************************************\
*                                                                *
*   Assign charges into ids using charge antialiasing method     *
*                                                                *
\****************************************************************/

void charge_antialias_method() {	
  int il,jl,kl,chk,idp,iidp,ngrds;
  double x,y,z,xi,yi,zi,rr,r2,rvdw2,sum;
  /* Define parameters */	
  double grdlng_nsubdiv = Grdlng / NSUBDIV;
  
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*                 Charge antialiasing method               *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  
  /* allocate memory */	
  int ndots[(NCAMNEIBORS*NCAMNEIBORS*NCAMNEIBORS)]; 
  
  /* Initialize the charge grid */
  for(int idp = 0; idp < NgrdxNgrdyNgrdz; idp++) {
    Grdchg[idp] = 0.0;
  }
  
   /********************************************\
   *                                            *
   *    Assign the protein charges into grids   *
   *                                            *
   \********************************************/

  for(int l = 0; l < Natms_tub; l++) {
    il = (int)floor(Atmx[l] * Igrdlng) + Ngrdx_2;
    jl = (int)floor(Atmy[l] * Igrdlng) + Ngrdy_2;
    kl = (int)floor(Atmz[l] * Igrdlng) + Ngrdz_2;
    r2 = (Atmrad[l] + Grdlng) * (Atmrad[l] + Grdlng);
    rvdw2 = Atmrad[l] * Atmrad[l];

    if(il < 0 || il > Ngrdx-1 || jl < 0 || jl > Ngrdy-1 || kl < 0 || kl > Ngrdz-1) {
      cout << "Error: Grid in charge antialias method exceeds the max grid" << endl; 
      exit(0);
    }
    
    /* Count the number of subdivisions within vdw radius for one grid point */
    int i;
    for(sum = 0.0, ngrds = 0, i = il-NCAMNEIBORS/2; i < il+NCAMNEIBORS/2; i++)
      if(i >= 0 && i < Ngrdx) 
	for(int j = jl-NCAMNEIBORS/2; j < jl+NCAMNEIBORS/2; j++) 
	  if(j >= 0 && j < Ngrdy) 
	    for(int k = kl-NCAMNEIBORS/2; k < kl+NCAMNEIBORS/2; k++) 
	      if(k >= 0 && k < Ngrdz) {
		iidp = (i-il+NCAMNEIBORS/2)*NCAMNEIBORS*NCAMNEIBORS + (j-jl+NCAMNEIBORS/2)*NCAMNEIBORS + k-kl+NCAMNEIBORS/2;
		ndots[iidp] = 0;
		x = (i - Ngrdx_2) * Grdlng - Atmx[l];
		y = (j - Ngrdy_2) * Grdlng - Atmy[l];
		z = (k - Ngrdz_2) * Grdlng - Atmz[l];
		rr = x*x + y*y + z*z;
		if(rr < r2)
		  chk = 0;
		for(int ii = 0 ; ii < NSUBDIV; ii++) for(int jj = 0; jj < NSUBDIV; jj++) for(int kk = 0; kk < NSUBDIV; kk++) {			
		  xi = x + (ii - 0.5 * (NSUBDIV - 1.0)) * grdlng_nsubdiv; 
		  yi = y + (jj - 0.5 * (NSUBDIV - 1.0)) * grdlng_nsubdiv; 
		  zi = z + (kk - 0.5 * (NSUBDIV - 1.0)) * grdlng_nsubdiv; 
		  rr = xi * xi + yi * yi + zi * zi;	
		  if(rr < rvdw2) {
		    ndots[iidp]++;
		    if(chk == 0) ngrds++;
		    chk = 1;
		  }
		}
		if(ndots[iidp] > 0) sum += ndots[iidp];
	      }
    /* end of i,j,k */
    /* Assign the charges into grid points */	
    if(ngrds >= NMINGRDS) {
      for(int i = il-NCAMNEIBORS/2; i < il+NCAMNEIBORS/2; i++) 
	if(i >= 0 && i < Ngrdx) 
	  for(int j = jl-NCAMNEIBORS/2; j < jl+NCAMNEIBORS/2; j++) 
	    if(j >= 0 && j < Ngrdy) 
	      for(int k = kl-NCAMNEIBORS/2; k < kl+NCAMNEIBORS/2; k++) 
		if(k >= 0 && k < Ngrdz) {
		  iidp = (i-il+NCAMNEIBORS/2)*NCAMNEIBORS*NCAMNEIBORS + (j-jl+NCAMNEIBORS/2)*NCAMNEIBORS + k-kl+NCAMNEIBORS/2;
		  if(ndots[iidp] > 0) {		
		    idp = i*Ngrdy*Ngrdz + j*Ngrdz + k;
		    Grdchg[idp] += Atmchg[l] * ndots[iidp] / sum;
		  }
		}
    }
    else if(ngrds < NMINGRDS) {
      for(int i = il; i < il + 2; i++) for(int j = jl; j < jl + 2; j++) for(int k = kl; k < kl + 2; k++) { 
	idp = i * NgrdyNgrdz + j * Ngrdz + k;
	x = 1.0 - fabs(i - Ngrdx_2 - Atmx[l] * Igrdlng);
	y = 1.0 - fabs(j - Ngrdy_2 - Atmy[l] * Igrdlng);
	z = 1.0 - fabs(k - Ngrdz_2 - Atmz[l] * Igrdlng);
	Grdchg[idp] += x * y * z * Atmchg[l]; 
      }
    } 
    /* finish the charge anti-alias method for protein */
  }
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*                  Charge assignment is done!!!            *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
}
 /****************************************************************\
 *                                                                *
 *	        Assign dielectric constants into grids            * 
 *                                                                *
 \****************************************************************/
 
void assign_dielects() {		
  double x,y,z,r2;
  /* Coarse assignment of dielectric constant */ 
  
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*     Begin Coarse assignment of dielectric constants      *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  
  /* Assign water dielectric constant into all grids */	
  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    
    /* For Epsx */	
    if(i < Ngrdx-1) {
      int idx = i*Ngrdy*Ngrdz + j*Ngrdz + k;
      Epsx[idx] = WATEPS;
      x = (i - Ngrdx/2.0 + 1.0) * Grdlng;
      y = (j - (Ngrdy-1.0)/2.0) * Grdlng;
      z = (k - (Ngrdz-1.0)/2.0) * Grdlng;
      r2 = x * x + y * y;	
      if(r2 > Tubrad_min2 && r2 < Gridrad_max2 && fabs(z) < Tublng_hlf) {
	Epsx[idx] = Memeps;
      }
      else if(r2 > Tubrad_min2 && r2 < Tubrad_max2 && fabs(z) < Tublng_hlf) {
	Epsx[idx] = Tubeps;
      }
    }
    /* For Epsy */	
    if(j < Ngrdy-1) {
      int idy = i*(Ngrdy-1)*Ngrdz + j*Ngrdz + k;
      Epsy[idy] = WATEPS;
      x = (i - (Ngrdx-1.0)/2.0) * Grdlng;
      y = (j - Ngrdy/2.0 + 1.0) * Grdlng;
      z = (k - (Ngrdz-1.0)/2.0) * Grdlng;
      r2 = x * x + y * y;	
      if(r2 > Tubrad_min2 && r2 < Gridrad_max2 && fabs(z) < Tublng_hlf) {
	Epsy[idy] = Memeps;
      }
      else if(r2 > Tubrad_min2 && r2 < Tubrad_max2 && fabs(z) < Tublng_hlf) {
	Epsy[idy] = Tubeps;
      }
    }	
    /* For Epsz */	
    if(k < Ngrdz-1) {
      int idz = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + k;
      Epsz[idz] = WATEPS;
      x = (i - (Ngrdx-1.0)/2.0) * Grdlng;
      y = (j - (Ngrdy-1.0)/2.0) * Grdlng;
      z = (k - Ngrdz/2.0 + 1.0) * Grdlng;
      r2 = x * x + y * y;
      if(r2 > Tubrad_min2 && r2 < Gridrad_max2 && fabs(z) < Tublng_hlf) {
	Epsz[idz] = Memeps;
      }
      else if(r2 > Tubrad_min2 && r2 < Tubrad_max2 && fabs(z) < Tublng_hlf) {
	Epsz[idz] = Tubeps;
      }
    }	
  }

   /************************************************************\	
   *                                                            *	
   *        Assign dielectric constants for channel water       *
   *                                                            *	
   \************************************************************/

  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    
    /* For Epsx */	
    int idx = i*NgrdyNgrdz + j*Ngrdz + k;
    if(i < Ngrdx-1 && Epsx[idx] == WATEPS) {
      x = (i - Ngrdx/2.0 + 1.0) * Grdlng;
      y = (j - (Ngrdy-1.0)/2.0) * Grdlng;
      z = (k - (Ngrdz-1.0)/2.0) * Grdlng;
      r2 = x * x + y * y;	
      if(r2 < Tubrad_min2 && fabs(z) < Tublng_hlf) Epsx[idx] = Chweps;
    }
    
    /* For Epsy */
    int idy = i*(Ngrdy-1)*Ngrdz + j*Ngrdz + k;
    if(j < Ngrdy-1 && Epsy[idy] == WATEPS) {
      x = (i - (Ngrdx-1.0)/2.0) * Grdlng;
      y = (j - Ngrdy/2.0 + 1.0) * Grdlng;
      z = (k - (Ngrdz-1.0)/2.0) * Grdlng;
      r2 = x * x + y * y;	
      if(r2 < Tubrad_min2 && fabs(z) < Tublng_hlf) Epsy[idy] = Chweps;
    }
    
    /* For Epsz */
    int idz = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + k;
    if(k < Ngrdz-1 && Epsz[idz] == WATEPS) {	
      x = (i - (Ngrdx-1.0)/2.0) * Grdlng;
      y = (j - (Ngrdy-1.0)/2.0) * Grdlng;
      z = (k - Ngrdz/2.0 + 1.0) * Grdlng;
      r2 = x * x + y * y;	
      if(r2 < Tubrad_min2 && fabs(z) < Tublng_hlf) Epsz[idz] = Chweps;
    }
  }
  
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*      Assignment of dielectric constants is done!!!       *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
}

/****************************************************************\
*                                                                *
*        Set up regions and Debye-Huckel parameters              *
*                                                                *
\****************************************************************/

void assign_regions() {

  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*       Set up regions and Debye-Huckel parameters         *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  
    /*****************************************\
    *                                         *
    * Set up integer Debye-Huckel parameters  *
    *                                         *
    \*****************************************/

  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {

    /* bulk water region */ 
    int idp = i*NgrdyNgrdz + j*Ngrdz + k;
    Grdid[idp] = 0;	
    
    /* Set up side wall regions */
    if(Flag_set_side_wall_boundary)
      if(i == 0 || i == Ngrdx-1 || j == 0 || j == Ngrdy-1) {
	Grdid[idp] = 5;
	continue;
      }           	
    
    double x = (i - Ngrdx_2) * Grdlng;
    double y = (j - Ngrdy_2) * Grdlng; 
    double z = (k - Ngrdz_2) * Grdlng; 
    double r2 = x * x + y * y;  
    
    /* Protein + ion exclusion region */
    if(r2 > Tubionrad_min2 && r2 < Tubionrad_max2 && fabs(z) < (Tublng_hlf + Ionrad)) {
      Grdid[idp] = 2;
    }
    /* Membrane region */
    if(r2 > Tubrad_min2 && r2 < Gridrad_max2 && fabs(z) < Tublng_hlf) {
      Grdid[idp] = 3;
    }
    /* Protein region */
    if(r2 > Tubrad_min2 && r2 < Tubrad_max2 && fabs(z) < Tublng_hlf) {
      Grdid[idp] = 1;
    }
    /* Channel water regions */	
    if(Grdid[idp] == 0 && fabs(z) < Tublng_hlf) {
      Grdid[idp] = 4;
    }
  }
  
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*   Setting up ion concentration and ion exclusion map     *" << endl
       << "*                       is done!!                          *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;	
}

/****************************************************************\
*                                                                *
*             Initialize the electrostatic potential             *
*               and set up the boundary condition                *
*                                                                *
\****************************************************************/

void assign_diffusion_const() {
  int n,idp,iidp; 

  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*               Set up the diffusion constants             *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  
  /* Assign the ionic concentration into grid using the method of Gilson et al. */
  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    idp = i*NgrdyNgrdz + j*Ngrdz + k;
    Cdfn[idp] = 0.0;
    Adfn[idp] = 0.0;
    
    /* in bulk water and channel water */
    if(Grdid[idp] == 0) {
      n = 0;
      for(int ii = i-1; ii < i+2; ii+=2) {
	if(ii == -1) iidp = (Ngrdx-2)*NgrdyNgrdz + j*Ngrdz + k;
	else if(ii == Ngrdx) iidp = 1*NgrdyNgrdz + j*Ngrdz + k;
	else iidp = ii*NgrdyNgrdz + j*Ngrdz + k;
	if(Grdid[iidp] == 4) n++;
      }
      
      for(int jj = j-1; jj < j+2; jj+=2) {
	if(jj == -1) iidp = i*NgrdyNgrdz + (Ngrdy-2)*Ngrdz + k;
	else if(jj == Ngrdy-1) iidp = i*NgrdyNgrdz + 1*Ngrdz + k;
	else iidp = i*NgrdyNgrdz + jj*Ngrdz + k;
	if(Grdid[iidp] == 4) n++;
      }
      for(int kk = k-1; kk < k+2; kk+=2) {
	if(kk == -1) iidp = i*NgrdyNgrdz + j*Ngrdz + (Ngrdz-2);
	else if(kk == Ngrdz-1) iidp = i*NgrdyNgrdz + j*Ngrdz + 1;
	else iidp = i*NgrdyNgrdz + j*Ngrdz + kk;
	if(Grdid[iidp] == 4) n++;
      }
      Cdfn[idp] = Catdfnblk * (6.0 - n) / 6.0 + Catdfnchn * n / 6.0;
      Adfn[idp] = Anidfnblk * (6.0 - n) / 6.0 + Anidfnchn * n / 6.0;	
    }
    /* for channel diffusion constant */
    else if(Grdid[idp] == 4) {
      n = 0;
      for(int ii = i-1; ii < i+2; ii+=2) {
	if(ii == -1) iidp = (Ngrdx-2)*NgrdyNgrdz + j*Ngrdz + k;
	else if(ii == Ngrdx) iidp = 1*NgrdyNgrdz + j*Ngrdz + k;
	else iidp = ii*NgrdyNgrdz + j*Ngrdz + k;
	if(Grdid[iidp] == 0) n++;
      }
      for(int jj = j-1; jj < j+2; jj+=2) {
	if(jj == -1) iidp = i*NgrdyNgrdz + (Ngrdy-2)*Ngrdz + k;
	else if(jj == Ngrdy-1) iidp = i*NgrdyNgrdz + 1*Ngrdz + k;
	else iidp = i*NgrdyNgrdz + jj*Ngrdz + k;
	if(Grdid[iidp] == 0) n++;
      }   
      for(int kk = k-1; kk < k+2; kk+=2) {
	if(kk == -1) iidp = i*NgrdyNgrdz + j*Ngrdz + (Ngrdz-2);
	else if(kk == Ngrdz-1) iidp = i*NgrdyNgrdz + j*Ngrdz + 1;
	else iidp = i*NgrdyNgrdz + j*Ngrdz + kk;
	if(Grdid[iidp] == 0) n++;
      }
      Cdfn[idp] = Catdfnchn * (6.0 - n) / 6.0 + Catdfnblk * n / 6.0;
      Adfn[idp] = Anidfnchn * (6.0 - n) / 6.0 + Anidfnblk * n / 6.0;
    }
  }
}

/****************************************************************\
*                                                                *
*             Initialize the electrostatic potential             *
*               and set up the boundary condition                *
*                                                                *
\****************************************************************/

void initialize(double voltg) {	
  double z,slp,anislp,anicnc,ratio;
  int idp;  

  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*            Initialize electrostatic potential            *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  
  slp = - EV2KCALPMOL * voltg / Tublng;	
  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    idp = i*NgrdyNgrdz + j*Ngrdz + k;
    z = (k - Ngrdz_2) * Grdlng; 
    /* boundary */	
    if(z > Tublng_hlf) Phi[idp] = 0.0; 
    else if(z < -Tublng_hlf) Phi[idp] = EV2KCALPMOL * voltg;
    else Phi[idp] = slp *  (z - Tublng_hlf);
  }
  
  cout << "************************************************************" << endl
       << "*                                                          *" << endl
       << "*        Initialize cation and anion concentrations        *" << endl
       << "*                                                          *" << endl
       << "************************************************************" << endl;
  cout << endl;
  
  /* Assign the ionic concentration into grid using the method of Gilson et al. */
  slp = (Topcnc - Botcnc) / Tublng;
  /* account for different ion valences
     this is simplified since only chlorine anions are used */
  anislp = Catchg*slp;  
  anicnc = Catchg*Botcnc;
  
  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    idp = i*NgrdyNgrdz + j*Ngrdz + k;
    z = (k - Ngrdz_2) * Grdlng; 
    
    /* in reservoir and in channel water */
    if(Grdid[idp] == 0 || Grdid[idp] == 4) {
      Ccnc[idp] = slp * (z + Tublng_hlf) + Botcnc;
      Acnc[idp] = anislp * (z + Tublng_hlf) + anicnc;
    }
    
    /* in protein, ion exclusion region, membrane, and side walls */	
    else if(Grdid[idp] == 1 || Grdid[idp] == 2 || Grdid[idp] == 3 || Grdid[idp] == 5){
      Ccnc[idp] = 0.0;
      Acnc[idp] = 0.0;
    }
  }
}

/****************************************************************\
*                                                                *
*             Calculate the potentials at each grid point        *
*                                                                *
\****************************************************************/

void calculate_electrostatic_potential_concentration() {

  int ip,idp,i_ini,i_fin,j_ini,j_fin,k_ini,k_fin;
  double xp,xm,yp,ym,zp,zm;
  double cxpu,cxpl,cxmu,cxml,cypu,cypl,cymu,cyml,czpu,czpl,czmu,czml;
  double axpl,axpu,axmu,axml,aypu,aypl,aymu,ayml,azpu,azpl,azmu,azml;
  double cnc,nmr,dnm;
  double tmp,tmp1,tmp2,tmp3;
  double cncnstgrdlng2,coulcnst_grdlng,err1,err2,err3,sum1,sum2,sum3;
  double omega1,omega2,rjac,resid;
  
  /* Define parameters */	
  cncnstgrdlng2 = CNCNST * Grdlng * Grdlng;
  coulcnst_grdlng = COULCNST / Grdlng;
  omega1 = OMEGA_INI;
  omega2 = OMEGA_INI;
  if(Flag_fixed_boundary) {
    rjac = (cos(M_PI / Ngrdx) + cos(M_PI / Ngrdy) + cos(M_PI / Ngrdz)) / 3.0;
  }
  else if(Flag_periodic_boundary) {
    rjac = (cos(2.0 * M_PI / Ngrdx) + cos(2.0 * M_PI / Ngrdy) + cos(2.0 * M_PI / Ngrdz)) / 3.0;
  }	

  cout << "spectral radius =    " << rjac << endl << endl;

  /* Allocate memory */	
  double* src = new double[NgrdxNgrdyNgrdz];
  double* eps6 = new double[NgrdxNgrdyNgrdz];
  int* edx1 = new int[NgrdxNgrdyNgrdz];
  int* edx2 = new int[NgrdxNgrdyNgrdz];	
  int* idx1 = new int[NgrdxNgrdyNgrdz];	
  int* idx2 = new int[NgrdxNgrdyNgrdz];
  int* edy1 = new int[NgrdxNgrdyNgrdz];
  int* edy2 = new int[NgrdxNgrdyNgrdz];	
  int* idy1 = new int[NgrdxNgrdyNgrdz];	
  int* idy2 = new int[NgrdxNgrdyNgrdz];	
  int* edz1 = new int[NgrdxNgrdyNgrdz];	
  int* edz2 = new int[NgrdxNgrdyNgrdz];	
  int* idz1 = new int[NgrdxNgrdyNgrdz];	
  int* idz2 = new int[NgrdxNgrdyNgrdz];	
  double* tmp4cx1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4ax1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4cx2 = new double[NgrdxNgrdyNgrdz];
  double* tmp4ax2 = new double[NgrdxNgrdyNgrdz];
  double* tmp4cy1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4ay1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4cy2 = new double[NgrdxNgrdyNgrdz];
  double* tmp4ay2 = new double[NgrdxNgrdyNgrdz];
  double* tmp4cz1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4az1 = new double[NgrdxNgrdyNgrdz];
  double* tmp4cz2 = new double[NgrdxNgrdyNgrdz];
  double* tmp4az2 = new double[NgrdxNgrdyNgrdz];
  
  /* prepare */
  for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
    idp = i*NgrdyNgrdz + j*Ngrdz + k; 	
    if(i == 0) {	
      edx1[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      edx2[idp] = (Ngrdx-2)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx1[idp] = (i+1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx2[idp] = (Ngrdx-2)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
    }	
    else if(i == Ngrdx-1) {
      edx1[idp] = 0*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      edx2[idp] = (i-1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx1[idp] = 1*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx2[idp] = (i-1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
    }
    else {
      edx1[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      edx2[idp] = (i-1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx1[idp] = (i+1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
      idx2[idp] = (i-1)*Ngrdy*Ngrdz + j*Ngrdz + k; 	
    }
    
    if(j == 0) {	
      edy1[idp] = i*(Ngrdy-1)*Ngrdz + j*Ngrdz + k;
      edy2[idp] = i*(Ngrdy-1)*Ngrdz + (Ngrdy-2)*Ngrdz + k;
      idy1[idp] = i*Ngrdy*Ngrdz + (j+1)*Ngrdz + k;
      idy2[idp] = i*Ngrdy*Ngrdz + (Ngrdy-2)*Ngrdz + k;
    }	
    else if(j == Ngrdy-1) {	
      edy1[idp] = i*(Ngrdy-1)*Ngrdz + 0*Ngrdz + k;
      edy2[idp] = i*(Ngrdy-1)*Ngrdz + (j-1)*Ngrdz + k;
      idy1[idp] = i*Ngrdy*Ngrdz + 1*Ngrdz + k;
      idy2[idp] = i*Ngrdy*Ngrdz + (j-1)*Ngrdz + k;
    }	
    else {	
      edy1[idp] = i*(Ngrdy-1)*Ngrdz + j*Ngrdz + k;
      edy2[idp] = i*(Ngrdy-1)*Ngrdz + (j-1)*Ngrdz + k;
      idy1[idp] = i*Ngrdy*Ngrdz + (j+1)*Ngrdz + k;
      idy2[idp] = i*Ngrdy*Ngrdz + (j-1)*Ngrdz + k;
    }
    
    if(k == 0) {	
      edz1[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + k; 	
      edz2[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + (Ngrdz-2);
      idz1[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + (k+1);
      idz2[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + (Ngrdz-2);
    }
    else if(k == Ngrdz-1) {
      edz1[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + 0; 	
      edz2[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + (k-1);
      idz1[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + 1;
      idz2[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + (k-1);
    }	
    else {	
      edz1[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + k; 	
      edz2[idp] = i*Ngrdy*(Ngrdz-1) + j*(Ngrdz-1) + (k-1);
      idz1[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + (k+1);
      idz2[idp] = i*Ngrdy*Ngrdz + j*Ngrdz + (k-1);
    }
    
    src[idp] = coulcnst_grdlng * Grdchg[idp];
    eps6[idp] = Epsx[edx1[idp]] + Epsx[edx2[idp]] + Epsy[edy1[idp]] + Epsy[edy2[idp]] + Epsz[edz1[idp]] + Epsz[edz2[idp]];
    
    tmp4cx1[idp] = 0.5 * (Cdfn[idx1[idp]] + Cdfn[idp]);
    tmp4ax1[idp] = 0.5 * (Adfn[idx1[idp]] + Adfn[idp]);
    tmp4cx2[idp] = 0.5 * (Cdfn[idx2[idp]] + Cdfn[idp]);
    tmp4ax2[idp] = 0.5 * (Adfn[idx2[idp]] + Adfn[idp]);
    
    tmp4cy1[idp] = 0.5 * (Cdfn[idy1[idp]] + Cdfn[idp]);
    tmp4ay1[idp] = 0.5 * (Adfn[idy1[idp]] + Adfn[idp]);
    tmp4cy2[idp] = 0.5 * (Cdfn[idy2[idp]] + Cdfn[idp]);
    tmp4ay2[idp] = 0.5 * (Adfn[idy2[idp]] + Adfn[idp]);
    
    tmp4cz1[idp] = 0.5 * (Cdfn[idz1[idp]] + Cdfn[idp]);
    tmp4az1[idp] = 0.5 * (Adfn[idz1[idp]] + Adfn[idp]);
    tmp4cz2[idp] = 0.5 * (Cdfn[idz2[idp]] + Cdfn[idp]);
    tmp4az2[idp] = 0.5 * (Adfn[idz2[idp]] + Adfn[idp]);    
  }

  /* Set up boundary condition */ 
  if(Flag_fixed_boundary) {	
    i_ini = 1; i_fin = Ngrdx - 1;
    j_ini = 1; j_fin = Ngrdy - 1;
    k_ini = 1; k_fin = Ngrdz - 1;
  }
  else if(Flag_periodic_boundary) {	
    i_ini = 0; i_fin = Ngrdx;
    j_ini = 0; j_fin = Ngrdy;
    k_ini = 1; k_fin = Ngrdz - 1;
  }

  cout << "*Tot_ese = total electrostatic energy" << endl; 
  
  /* Begin iteration */	
  for(int itr = 0; ; itr++) {
    /* Calculate the electrostatic potentials */	
    if(Flag_pot_calc) {
      for(err1 = sum1 = Tot_ese = 0.0, ip = 0; ip < 2; ip++) { 
	for(int i = i_ini; i < i_fin; i++) for(int j = j_ini; j < j_fin; j++) for(int k = k_ini + (i+j+ip) % 2; k < k_fin; k+=2) {
	  idp = i*NgrdyNgrdz + j*Ngrdz + k; 	
	  
	  xp = Epsx[edx1[idp]] * Phi[idx1[idp]];
	  xm = Epsx[edx2[idp]] * Phi[idx2[idp]];
	  
	  yp = Epsy[edy1[idp]] * Phi[idy1[idp]];
	  ym = Epsy[edy2[idp]] * Phi[idy2[idp]];
	  
	  zp = Epsz[edz1[idp]] * Phi[idz1[idp]];
	  zm = Epsz[edz2[idp]] * Phi[idz2[idp]];
	  
	  cnc = cncnstgrdlng2 * (Catchg * Ccnc[idp] + Anichg * Acnc[idp]);
	  nmr = xp + xm + yp + ym + zp + zm + src[idp] + cnc; 
	  dnm = eps6[idp];
	  
	  sum1 += fabs(Phi[idp]);	
	  resid = Phi[idp] - nmr / dnm;
	  Phi[idp] -= omega1 * resid;	
	  err1 += fabs(resid);	
	  Tot_ese += 0.5 * Grdchg[idp] * Phi[idp];
	} 
	/*end of i,j,k */
	//omega1 = (itr == 0 && ip == 0 ? 1.0 / (1.0 - 0.5 * rjac * rjac) : 1.0 / (1.0 - 0.25 * rjac * rjac * omega1));	
      } 
      /* end of ip */
      err1 /= sum1;
      if(!Flag_cnc_calc && itr % DSPSTP == 0) {
	cout << myid << " : " << itr << "  : err1    = " << err1 << endl;	
	//cout << itr << "  : omega1  = " << omega1 << endl;
	cout << "    " << itr << "  : Tot_ese = " << Tot_ese << " (Kcal/mol)" << endl;
	cout << "    " << itr << "  : Tot_ese = " << (Tot_ese / KBT) << " (KT)" << endl;
      }
    } /* end electrostatic potential calculation */
    /* Calculate cation and anion concentrations */	
    if(Flag_cnc_calc) {
      for(err2 = sum2 = err3 = sum3 = 0.0, ip = 0; ip < 2; ip++) { 
	for(int i = i_ini; i < i_fin; i++) for(int j = j_ini; j < j_fin; j++) for(int k = k_ini + (i+j+ip) % 2; k < k_fin; k+=2) {
	  idp = i*NgrdyNgrdz + j*Ngrdz + k; 	
	  if(Grdid[idp] == 4 || Grdid[idp] == 0) {
	    /* cation czoncentration */	
	    if(Grdid[idx1[idp]] == 4 || Grdid[idx1[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idx1[idp]] - Phi[idp]);
	      cxpu = (1.0 + tmp3) * tmp4cx1[idp];
	      cxpl = (1.0 - tmp3) * tmp4cx1[idp];
	      axpu = (1.0 - tmp3) * tmp4ax1[idp];
	      axpl = (1.0 + tmp3) * tmp4ax1[idp];
	    }
	    else cxpl = cxpu = axpu = axpl = 0.0;
	    if(Grdid[idx2[idp]] == 4 || Grdid[idx2[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idx2[idp]] - Phi[idp]);
	      cxmu = (1.0 + tmp3) * tmp4cx2[idp];
	      cxml = (1.0 - tmp3) * tmp4cx2[idp];
	      axmu = (1.0 - tmp3) * tmp4ax2[idp];
	      axml = (1.0 + tmp3) * tmp4ax2[idp];
	    }
	    else cxml = cxmu = axmu = axml = 0.0;
	    if(Grdid[idy1[idp]] == 4 || Grdid[idy1[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idy1[idp]] - Phi[idp]);
	      cypu = (1.0 + tmp3) * tmp4cy1[idp];
	      cypl = (1.0 - tmp3) * tmp4cy1[idp];
	      aypu = (1.0 - tmp3) * tmp4ay1[idp];
	      aypl = (1.0 + tmp3) * tmp4ay1[idp];
	    }
	    else cypl = cypu = aypu = aypl = 0.0;
	    if(Grdid[idy2[idp]] == 4 || Grdid[idy2[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idy2[idp]] - Phi[idp]);
	      cymu = (1.0 + tmp3) * tmp4cy2[idp];
	      cyml = (1.0 - tmp3) * tmp4cy2[idp];
	      aymu = (1.0 - tmp3) * tmp4ay2[idp];
	      ayml = (1.0 + tmp3) * tmp4ay2[idp];
	    }
	    else cyml = cymu = aymu = ayml = 0.0;
	    if(Grdid[idz1[idp]] == 4 || Grdid[idz1[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idz1[idp]] - Phi[idp]);
	      czpu = (1.0 + tmp3) * tmp4cz1[idp];
	      czpl = (1.0 - tmp3) * tmp4cz1[idp];
	      azpu = (1.0 - tmp3) * tmp4az1[idp];
	      azpl = (1.0 + tmp3) * tmp4az1[idp];
	    }
	    else czpl = czpu = azpu = azpl = 0.0;
	    if(Grdid[idz2[idp]] == 4 || Grdid[idz2[idp]] == 0) {
	      tmp3 = 0.5 * Ikbt * (Phi[idz2[idp]] - Phi[idp]);
	      czmu = (1.0 + tmp3) * tmp4cz2[idp];
	      czml = (1.0 - tmp3) * tmp4cz2[idp];
	      azmu = (1.0 - tmp3) * tmp4az2[idp];
	      azml = (1.0 + tmp3) * tmp4az2[idp];
	    }	
	    else czml = czmu = azmu = azml = 0.0;
	    /* cation concentration */
	    nmr = cxpu * Ccnc[idx1[idp]] + cxmu * Ccnc[idx2[idp]]
	      +cypu * Ccnc[idy1[idp]] + cymu * Ccnc[idy2[idp]]
	      +czpu * Ccnc[idz1[idp]] + czmu * Ccnc[idz2[idp]];
	    dnm = cxpl + cxml + cypl + cyml + czpl + czml;
	    
	    sum2 += fabs(Ccnc[idp]);	
	    resid = Ccnc[idp] - nmr / dnm;
	    Ccnc[idp] -= omega2 * resid;
	    err2 += fabs(resid);
	    
	    /* anion concentration */
	    nmr = axpu * Acnc[idx1[idp]] + axmu * Acnc[idx2[idp]]
	      +aypu * Acnc[idy1[idp]] + aymu * Acnc[idy2[idp]]
	      +azpu * Acnc[idz1[idp]] + azmu * Acnc[idz2[idp]];
	    dnm = axpl + axml + aypl + ayml + azpl + azml;
	    
	    sum3 += fabs(Acnc[idp]);	
	    resid = Acnc[idp] - nmr / dnm;
	    Acnc[idp] -= omega2 * resid;
	    err3 += fabs(resid);
	  }
	}
	/* end of i,j,k */
	//omega2 = (itr == 0 && ip == 0 ? 1.0 / (1.0 - 0.5 * rjac * rjac) : 1.0 / (1.0 - 0.25 * rjac * rjac * omega2));	
      }
      /* end of ip */
      err2 /= sum2;
      err3 /= sum3;
      if(!Flag_pot_calc && itr % DSPSTP == 0) {
	cout << itr << "  : err2   =   " << err2 << endl;	
	cout << itr << "  : err3   =   " << err3 << endl;	
	//cout << itr << "  : omega2 =   " << omega2 << endl;	
      }
     } 
    /* End of concentration calculation */
    if(Flag_cnc_calc && Flag_pot_calc && itr % DSPSTP == 0) {
      cout << myid << " : " << itr << "  : err1 = " << err1 << "  : err2 = " << err2 << "  : err3 = " << err3 << endl;
      //cout << itr << "  : omega1 = " << omega1 << " : omega2 = " << omega2 << endl;
      cout << "    " << itr << "  : Tot_ese = " << Tot_ese << " (Kcal/mol)" << endl;
      cout << "    " << itr << "  : Tot_ese = " << (Tot_ese / KBT) << " (KT)" << endl;
      cout << endl;
    }
    if(err1 < Tol1 && err2 < Tol2 && err3 < Tol2) {
      cout << "process " << myid << " : itr = " << itr << "  : SOLUTION FOUND!" << endl << endl;	
      break;
    }
    /* Account for instability of the calculation, abort all calculations if any one is unstable */
    if(isnan(err1) == 1 || isnan(err2) == 1 || isnan(err3) == 1){
      Calc_unstable = TRUE;
      cout << "process " << myid << " : itr = " << itr << "  : CALCULATION UNSTABLE! ABORTED" << endl << endl;
      break;
    }
  } 
  /* End of iteration */  
  /*Calculation of  Cflx and Aflx */
  if(!Calc_unstable){
    for(int i = 1; i < Ngrdx-1; i++) for(int j = 1; j < Ngrdy-1; j++) for(int k = 1; k < Ngrdz-1; k++) {
      
      idp = i*NgrdyNgrdz + j*Ngrdz + k;
      if(Grdid[idz1[idp]] == 4 || Grdid[idz1[idp]] == 0) {
	tmp1 = (Ccnc[idz1[idp]] - Ccnc[idp]) * Igrdlng;
	tmp2 = 0.5 * (Ccnc[idz1[idp]] + Ccnc[idp]);
	tmp3 = (Phi[idz1[idp]] - Phi[idp]) * Igrdlng;
	tmp  = (tmp1 + Ikbt * Catchg * tmp2 * tmp3) * tmp4cz1[idp];
      }	
      else tmp = 0.0;
      Cflx[idp] = -CRTCNST * tmp;
      
      if(Grdid[idz1[idp]] == 4 || Grdid[idz1[idp]] == 0) {	
	tmp1 = (Acnc[idz1[idp]] - Acnc[idp]) * Igrdlng;
	tmp2 = 0.5 * (Acnc[idz1[idp]] + Acnc[idp]);
	tmp3 = (Phi[idz1[idp]] - Phi[idp]) * Igrdlng;
	tmp  = (tmp1 + Ikbt * Anichg * tmp2 * tmp3) * tmp4az1[idp];
      }
      else tmp = 0.0;	
      Aflx[idp] = -CRTCNST * tmp;
    }
  }

  delete[] src;
  delete[] eps6;
  delete[] edx1;
  delete[] edx2;
  delete[] idx1;
  delete[] idx2;
  delete[] edy1;
  delete[] edy2;
  delete[] idy1;
  delete[] idy2;
  delete[] edz1;
  delete[] edz2;
  delete[] idz1;
  delete[] idz2;
}

/**********************************************************\
*                                                          *    
*                           Main                           *    
*                                                          *    
\**********************************************************/

int main(int argc, char* argv[]) {  
  
  /* local variables for iterations */
  int i,j,k,l,idp;
  double x,y,z,r,r2,sum,sum1,temp,cmax,amax,pmin;
  
  /* applied voltage parameters */
  double inivtg,dvol,voltg,*voltgsend;
  int max_steps;
  
  /* grid input related strings */
  string grid;
  string coarse( "coarse" );
  string fine( "fine" );
  string very( "very_fine" );
  
  /* ion input related strings */
  string ion;
  string Na( "Na" );
  string K( "K" );
  string Mg( "Mg" );
  string Ca( "Ca" );
  string Cl( "Cl" );
  
  /* output related strings and variables */
  flag zy_profile  =  TRUE;
  flag xy_profile  =  FALSE;
  flag xyz_profile =  FALSE;
  string profile;
  string yes( "yes" );
  string no( "no" );
  double zcut,xcut;
  
  /* sstreams for writing XML */
  stringstream phipath;
  stringstream catpath;
  stringstream anipath;
  stringstream voltg_label;
  stringstream datapoint;
  stringstream cat_val;
  stringstream ani_val;
  stringstream phi_val;
  stringstream error;

  /* MPI variables */
  int numprocs;
  double inputs[24];
  /****************************************************************\
  *                                                                *
  *                          Initialize                            *
  *                                                                *
  \****************************************************************/
  
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  MPI_Status status;
  
  /* declare Rappture library */
  RpLibrary* lib = NULL;
  
  if(myid == 0) {
    cout << "************************************************************" << endl
	 << "*                                                          *" << endl
	 << "*                    Reading driver file                   *" << endl
	 << "*                                                          *" << endl
	 << "************************************************************" << endl;
    
    /* define Rappture library */
    string filepath = (string)argv[1];
    lib = new RpLibrary(filepath);
    
    /* define system parameters */
    inputs[0]  = lib->getDouble("input.(inputs).(system).(Nrings).current");
    inputs[1]  = lib->getDouble("input.(inputs).(system).(Tubrad_min).current");
    inputs[2]  = lib->getDouble("input.(inputs).(system).(Tubcatchg).current");    
    inputs[3]  = lib->getDouble("input.(inputs).(system).(Tubanichg).current");
    inputs[4]  = lib->getDouble("input.(inputs).(system).(TubDplLng).current"); 
    inputs[5]  = lib->getDouble("input.(inputs).(system).(Tubatmrad).current");
    inputs[6]  = lib->getDouble("input.(inputs).(system).(Tubeps).current");
    inputs[7]  = lib->getDouble("input.(inputs).(system).(Memeps).current"); 
    
    /* define buffer parameters */
    inputs[8]  = lib->getDouble("input.(inputs).(ion).(buffer).(Topcnc).current");
    inputs[9]  = lib->getDouble("input.(inputs).(ion).(buffer).(Botcnc).current");
    
    /* define cation parameters */
    ion        = lib->getString("input.(inputs).(ion).group(cation).choice(cation).current");
    inputs[10] = lib->getDouble("input.(inputs).(ion).(cation).(Catdfnblk).current");
    inputs[11] = lib->getDouble("input.(inputs).(ion).(cation).(Catdfnchn).current");
    if(ion == Na) {
      inputs[12] = 1.02;
      inputs[13] = 1.0;
    }
    else if(ion == K) {
      inputs[12] = 1.38;
      inputs[13] = 1.0;
    }
    else if(ion == Mg) {
      inputs[12] = 0.72;
      inputs[13] = 2.0;
    }
    else if(ion == Ca) {
      inputs[12] = 0.99;
      inputs[13] = 2.0;
    }
    else {
      inputs[12] = lib->getDouble("input.(inputs).(ion).(cation).(Ionrad).current");
      inputs[13] = lib->getDouble("input.(inputs).(ion).(cation).(Catchg).current");
    }
    
    /* define anion parameters */
    ion        = lib->getString("input.(inputs).(ion).group(anion).choice(anion).current");
    inputs[14] = lib->getDouble("input.(inputs).(ion).(anion).(Anidfnblk).current");
    inputs[15] = lib->getDouble("input.(inputs).(ion).(anion).(Anidfnchn).current");
    if(ion == Cl) {
      inputs[16] = -1.0;
    }  
    
    /* define applied voltage */
    inivtg    = lib->getDouble("input.(inputs).(voltg).(inivtg).current");
    max_steps = numprocs;
    dvol      = lib->getDouble("input.(inputs).(voltg).(dvol).current");
    
    /* define numerical parameters */
    grid = lib->getString("input.(inputs).(itr).(grid).current"); 
    if(grid == coarse) {
      inputs[17] =  31; 
      inputs[18] =  31; 
      inputs[19] =  41;
      inputs[20] =   2.0;
    }
    else if(grid == fine) {
      inputs[17] =  61;
      inputs[18] =  61;
      inputs[19] =  81;
      inputs[20] =   1.0;
    }
    else if(grid == very) {
      inputs[17] =  121;
      inputs[18] =  121;
      inputs[19] =  161;
      inputs[20] =    0.5;
    }
    
    inputs[21] = lib->getDouble("input.(inputs).(itr).(Tol1).current");
    inputs[22] = lib->getDouble("input.(inputs).(itr).(Tol2).current");
    
    /* Define water parameters */
    inputs[23] = lib->getDouble("input.(inputs).(system).(Chweps).current");
    
    /* Define outputs */
    profile = lib->getString("input.(inputs).(outputs).(zy).current");
    if(profile == no) {
      zy_profile = FALSE;
    }
    xcut = lib->getDouble("input.(inputs).(outputs).(xcut).current");
    profile = lib->getString("input.(inputs).(outputs).(xy).current");
    if(profile == yes) {
      xy_profile = TRUE;
    }
    zcut = lib->getDouble("input.(inputs).(outputs).(zcut).current");
    profile = lib->getString("input.(inputs).(outputs).(3d).current");
    if(profile == yes) {
      xyz_profile = TRUE;
    }
    
     /****************************************************************\
     *                                                                *
     *                           Set up slaves                        *
     *                                                                *
     \****************************************************************/
    cout << "************************************************************" << endl
	 << "*                                                          *" << endl
	 << "*                     Setting up inputs                    *" << endl
	 << "*                                                          *" << endl
	 << "************************************************************" << endl;
  }
  
  MPI_Bcast(inputs, 24, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  Nrings     = (int)inputs[0]; 
  Tublng     = 5*(Nrings-1);
  Tubrad_min = inputs[1];
  Tubcatchg  = inputs[2]; 
  Tubanichg  = inputs[3];
  TubDplLng  = inputs[4];
  Tubrad_max = Tubrad_min + TubDplLng + 4.0;
  Tubatmrad  = inputs[5];
  Tubeps     = inputs[6];
  Memeps     = inputs[7];
  Topcnc     = inputs[8];
  Botcnc     = inputs[9];
  Catdfnblk  = inputs[10];
  Catdfnchn  = inputs[11];
  Ionrad     = inputs[12];
  Catchg     = inputs[13];
  Anidfnblk  = inputs[14];
  Anidfnchn  = inputs[15];
  Anichg     = inputs[16];
  Ngrdx      = (int)inputs[17];
  Ngrdy      = (int)inputs[18];
  Ngrdz      = (int)inputs[19];
  Grdlng     = inputs[20];
  Tol1       = inputs[21];
  Tol2       = inputs[22];
  Chweps     = inputs[23];

  if(myid == 0) {
    cout << "Nrings = " << Nrings << endl
	 << "Tublng = " << Tublng << endl
	 << "Tubrad_min = " << Tubrad_min << endl
	 << "Tubcatchg = " << Tubcatchg << endl
	 << "Tubanichg = " << Tubanichg << endl
	 << "TubDplLng = " << TubDplLng << endl
	 << "Tubrad_max = " << Tubrad_max << endl
	 << "Tubatmrad = " << Tubatmrad << endl
	 << "Tubeps = " << Tubeps << endl
	 << "Memeps = " << Memeps << endl
	 << "Topcnc = " << Topcnc << endl
	 << "Botcnc = " << Botcnc << endl
	 << "Catdfnblk = " << Catdfnblk << endl
	 << "Catdfnchn = " << Catdfnchn << endl
	 << "Ionrad = " << Ionrad << endl
	 << "Catchg = " << Catchg << endl
	 << "Anidfnblk = " << Anidfnblk << endl
	 << "Anidfnchn = " << Anidfnchn << endl
	 << "Anichg = " << Anichg << endl
	 << "inivtg = " << inivtg << endl
	 << "max_steps = " << max_steps << endl
	 << "dvol = " << dvol << endl
	 << "Ngrdx = " << Ngrdx << endl
	 << "Ngrdy = " << Ngrdy << endl
	 << "Ngrdz = " << Ngrdz << endl
	 << "Grdlng = " << Grdlng << endl
	 << "Tol1 = " << Tol1 << endl
	 << "Tol2 = " << Tol2 << endl
	 << "Chweps = " << Chweps << endl;

    voltgsend = new double[numprocs];
  }

    /****************************************************************\
    *                                                                *
    *        Set up external parameters and memory allocation        *
    *                                                                *
    \****************************************************************/

  setup_params_memalloc();
  
  /* allocate memory for output arrays */
  double* ccnc_avg = new double[Ngrdz];
  double* acnc_avg = new double[Ngrdz];
  double* cflx_avg = new double[Ngrdz];
  double* aflx_avg = new double[Ngrdz];
  double* ccrt_avg = new double[Ngrdz];
  double* acrt_avg = new double[Ngrdz];
  
  /* allocate output arrays for data gathering */
  double* ccnc_avg_master = new double[numprocs*Ngrdz];
  double* acnc_avg_master = new double[numprocs*Ngrdz];
  double* cflx_avg_master = new double[numprocs*Ngrdz];
  double* aflx_avg_master = new double[numprocs*Ngrdz];
  double* ccrt_avg_master = new double[numprocs*Ngrdz];
  double* acrt_avg_master = new double[numprocs*Ngrdz];
  double** phi_master = new double*[NgrdxNgrdy];
  double** ccnc_master = new double*[NgrdxNgrdy];
  double** acnc_master = new double*[NgrdxNgrdy];
  for(int ij = 0; ij < NgrdxNgrdy; ij++) {
    phi_master[ij] = new double[numprocs*Ngrdz];
    ccnc_master[ij] = new double[numprocs*Ngrdz];
    acnc_master[ij] = new double[numprocs*Ngrdz];
  }
    /**********************************************************\
    *                                                          *
    *     Read the coordinates of nanotube from minimized      *
    *     NAMD output file                                     * 
    *                                                          *
    \**********************************************************/
  
   read_pdb_file();
   
    /****************************************************************\
    *                                                                *
    *        Read "top_all27_prot_lipid.inp" and                     *
    *        "par_all27_prot_lipid.inp" to obtain atomic charges     *
    *        and VDW radius.                                         *
    *                                                                *
    \****************************************************************/
   if(myid == 0) {
     read_charges_vdwradii(); 
     
      /****************************************************************\
      *                                                                *
      *            assign the charges constants into grids             *
      *                                                                *
      \****************************************************************/

     charge_antialias_method(); 
     
     for(sum = 0.0,l = 0; l < Natms_tub; l++) {
       sum += Atmchg[l];
     }
     cout << "Total charge of protein =      " << sum << endl;
     
     for(sum = 0.0, idp = 0; idp < NgrdxNgrdyNgrdz; idp++) {
       sum += Grdchg[idp];
     } 
     cout << "Total charge of grid    =  " << sum << endl;
     
      /****************************************************************\
      *                                                                *
      *            assign the dielectric constants into grids          *
      *            by using Lee and Richards algorithm.                *
      *                                                                *
      \****************************************************************/
   
     assign_dielects();
     
      /****************************************************************\
      *                                                                *
      *  Set up ion concentration and ion exclusion map (Stern layer). *
      *                                                                *
      \****************************************************************/
	
     assign_regions();	
     assign_diffusion_const();
   
	 /****************************************************************\
	 *                                                                *
	 *             Calculate water accessible volume                  *
	 *                                                                *
         \****************************************************************/
 
     for(sum = sum1 = 0.0, i = 0; i < Ngrdx; i++) for(j = 0; j < Ngrdy; j++) for(k = 0; k < Ngrdz; k++) {
       idp = i*NgrdyNgrdz + j*Ngrdz + k;
       x = (i - Ngrdx_2) * Grdlng;
       y = (j - Ngrdy_2) * Grdlng; 
       z = (k - Ngrdz_2) * Grdlng; 
       r2 = x * x + y * y;
       r = sqrt(r2);
       
       /* calculate the ion accessible volume in z direction */   
       if(r2 <= Tubionrad_min2 && (Grdid[idp] == 0 || Grdid[idp] == 4)) Nvolz[k]++;
       
       /* calculate the ion accessible volume in r direction */
       l = (int)floor(r * Igrdlng);
       if(l < Ngrdx_2 && fabs(z) < Tublng_hlf  && (Grdid[idp] == 0 || Grdid[idp] == 4 || Grdid[idp] == 2)) Nvolr[l]++;	
       
       /* Calculate ion-available total volume of box */
       if((Grdid[idp] == 0 || Grdid[idp] == 4) && i < Ngrdx && j < Ngrdy && k < Ngrdz) sum++;
       if(Grdid[idp] == 4 && i < Ngrdx && j < Ngrdy && k < Ngrdz) sum1++;
     }
     
     /* calculate the ion accessible volume */   
     Watvol = sum * Grdvol; 
     Tub_watvol = sum1 * Grdvol; 
     
       /****************************************************************\
       *                                                                *
       *             Initialize the electrostatic potential             *
       *               and set up the boundary condition                *
       *                                                                *
       \****************************************************************/
  
  
     for(int step = 0; step < max_steps; step++) {
       voltg = inivtg + (step*dvol);
       cout << voltg << "V sent to process " << step << endl;
       voltgsend[step] = voltg;
     }
   }
   /* broadcast the initialized arrays */
   MPI_Bcast(Grdid, NgrdxNgrdyNgrdz, MPI_INT, 0, MPI_COMM_WORLD);
   MPI_Bcast(Grdchg, NgrdxNgrdyNgrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Epsx, ((Ngrdx-1)*Ngrdy*Ngrdz), MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Epsy, (Ngrdx*(Ngrdy-1)*Ngrdz), MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Epsz, (Ngrdx*Ngrdy*(Ngrdz-1)), MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Cdfn, NgrdxNgrdyNgrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Adfn, NgrdxNgrdyNgrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Ccnc, NgrdxNgrdyNgrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Acnc, NgrdxNgrdyNgrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   MPI_Bcast(Nvolz, Ngrdz, MPI_INT, 0, MPI_COMM_WORLD);
   /* send a unique voltage value */
   MPI_Scatter(voltgsend, 1, MPI_DOUBLE, &voltg, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
   /* initialize the Phi array */
   initialize(voltg);
     
    /****************************************************************\
    *                                                                *
    *             Calculate the potentials at each grid point        *
    *                                                                *
    \****************************************************************/
     cout << "************************************************************" << endl
	  << "*                        Process " << myid << "                         *" << endl
	  << "*            Calculating Electrostatic Potential           *" << endl
	  << "*                                                          *" << endl
	  << "************************************************************" << endl;
   

   calculate_electrostatic_potential_concentration();
	
    /****************************************************************\
    *                                                                *
    *                       Process the output                       *
    *                                                                *
    \****************************************************************/

   if(!Calc_unstable){
     /* Calculate the average concentrations */
     cout << "************************************************************" << endl
	  << "*                        Process " << myid << "                         *" << endl
	  << "*              Calculating Ion Concentrations              *" << endl
	  << "*                                                          *" << endl
	  << "************************************************************" << endl;
     
     for(int k = 0; k < Ngrdz; k++){ 
       int i,j;
       for(ccnc_avg[k] = acnc_avg[k] = 0.0, j = 0; j < Ngrdy; j++) for(i = 0; i < Ngrdx; i++) {	      
	 idp = i*NgrdyNgrdz + j*Ngrdz + k;
	 x = (i - Ngrdx_2) * Grdlng;
	 y = (j - Ngrdy_2) * Grdlng;
	 r2 = x * x + y * y;
	 
	 if(r2 <= Tubionrad_min2) {
	   ccnc_avg[k] += Ccnc[idp];
	   acnc_avg[k] += Acnc[idp];
	 }
       }
       ccnc_avg[k] /= Nvolz[k];	
       acnc_avg[k] /= Nvolz[k];
     }
     
     /* Calculation of ion currents */
     
     cout << "************************************************************" << endl
	  << "*                          Process " << myid << "                       *" << endl
	  << "*           Calculating Ion Flux and Ion Current           *" << endl
	  << "*                                                          *" << endl
	  << "************************************************************" << endl;
     
     for(int k = 0; k < Ngrdz; k++) {
       int i,j;
       for(cflx_avg[k] = aflx_avg[k] = ccrt_avg[k] = acrt_avg[k] = 0.0, i = 1; i < Ngrdx-1; i++)
	 for(j = 1; j < Ngrdy-1; j++) {
	   idp = i*NgrdyNgrdz + j*Ngrdz + k;
	   x = (i - Ngrdx_2) * Grdlng;
	   y = (j - Ngrdy_2) * Grdlng;
	   r2 = x * x + y * y;
	   
	   /* Average flux */	
	   if(r2 <= Tubionrad_min2) {
	     cflx_avg[k] += Cflx[idp];
	     aflx_avg[k] += Aflx[idp];
	   }
	   /* Average current */	
	   if(Grdid[idp] == 4 || Grdid[idp] == 0) {
	     ccrt_avg[k] += Cflx[idp];
	     acrt_avg[k] += Aflx[idp];
	   }
	 }
       cflx_avg[k] /= Nvolz[k];
       aflx_avg[k] /= Nvolz[k];
       
       ccrt_avg[k] *= Grdlng2;
       acrt_avg[k] *= Grdlng2;
     }
     
      /****************************************************************\
      *                                                                *
      *                    Collect data from slaves                    *
      *                                                                *
      \****************************************************************/
     /* sort pseudo 3d arrays into 2d arrays */
     for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
       idp = i*NgrdyNgrdz + j*Ngrdz + k;
       int iidp = i*Ngrdy + j;
       
       Phi2d[iidp][k] = Phi[idp]/EV2KCALPMOL;  /* convert units to V from kcal/mol/e */
       Ccnc2d[iidp][k] = Ccnc[idp];
       Acnc2d[iidp][k] = Acnc[idp];
     }
     
     /* Gather arrays using "stacking" to minimize each data transfer */
     MPI_Gather(ccrt_avg, Ngrdz, MPI_DOUBLE, ccrt_avg_master, Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
     MPI_Gather(acrt_avg, Ngrdz, MPI_DOUBLE, acrt_avg_master, Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
     MPI_Gather(ccnc_avg, Ngrdz, MPI_DOUBLE, ccnc_avg_master, Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
     MPI_Gather(acnc_avg, Ngrdz, MPI_DOUBLE, acnc_avg_master, Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
     for(int ij = 0; ij < NgrdxNgrdy; ij++) {
       MPI_Gather(Phi2d[ij], Ngrdz, MPI_DOUBLE, phi_master[ij], Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
       MPI_Gather(Ccnc2d[ij], Ngrdz, MPI_DOUBLE, ccnc_master[ij], Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
       MPI_Gather(Acnc2d[ij], Ngrdz, MPI_DOUBLE, acnc_master[ij], Ngrdz, MPI_DOUBLE, 0, MPI_COMM_WORLD);
     }
   }
   /****************************************************************\
   *                                                                *
   *                 Use master to create run file                  *
   *                                                                *
   \****************************************************************/

   if(myid == 0) {
     if(!Calc_unstable){
       /* iterate over each voltage value and write to XML */
       for(int step = 0; step < max_steps; step++) {
	 /* note that step is the same as myid and that max_steps = numprocs */
	 voltg = inivtg + (step*dvol);
	 
	 /* If an appreciable number of voltage points are used, print the current/voltage graph */
	 /* else just output the values as a string */
	 if(max_steps > 1) {
	   /* write the graph XML on step 0 */
	   if(step == 0) {
	     /* Cation Current */
	     lib->put("output.curve(ccrtvltg).about.group", "Current vs Voltage");
	     lib->put("output.curve(ccrtvltg).about.label", "Cation Current");
	     lib->put("output.curve(ccrtvltg).xaxis.label", "Applied Voltgage");
	     lib->put("output.curve(ccrtvltg).xaxis.units", "V");
	     lib->put("output.curve(ccrtvltg).yaxis.label", "Current");
	     lib->put("output.curve(ccrtvltg).yaxis.units", "pA");
	     
	     /* Anion Current */
	     lib->put("output.curve(acrtvltg).about.group", "Current vs Voltage");     
	     lib->put("output.curve(acrtvltg).about.label", "Anion Current");
	     lib->put("output.curve(acrtvltg).xaxis.label", "Applied Voltgage");
	     lib->put("output.curve(acrtvltg).xaxis.units", "V");
	     lib->put("output.curve(acrtvltg).yaxis.label", "Current");
	     lib->put("output.curve(acrtvltg).yaxis.units", "pA");
	   }
	   
	   cout << "************************************************************" << endl
		<< "*                                                          *" << endl
		<< "*             Writing Ion Current and Voltage              *" << endl
		<< "*                                                          *" << endl
		<< "************************************************************" << endl;
	   
	   int kk = Ngrdz_2 + (step*Ngrdz);
	   /* Cation Current at z = 0 */
	   cat_val << voltg << " " << ccrt_avg_master[kk] << endl;
	   lib->put("output.curve(ccrtvltg).component.xy", cat_val.str(), "", 1);
	   cat_val.str("");
	   
	   /* Anion Current at z = 0 */
	   ani_val << voltg << " " << -acrt_avg_master[kk] << endl;
	   lib->put("output.curve(acrtvltg).component.xy", ani_val.str(), "", 1);
	   ani_val.str("");
	 }
	 else if(max_steps == 1) {
	   /* if only one voltage is used, just return the point, not a graph */
	   lib->put("output.string(crtvltg).about.label", "Ion Current");
	   
	   cat_val << "Cation Current = " << ccrt_avg_master[Ngrdz_2] << " pA" << endl;
	   lib->put("output.string(crtvltg).current", cat_val.str(), "", 1);
	   cat_val.str("");
	   
	   ani_val << "Anion Current = " << -acrt_avg_master[Ngrdz_2] << " pA" << endl;
	   lib->put("output.string(crtvltg).current", ani_val.str(), "", 1);
	   ani_val.str("");
	 }
	 /* Write the XML for each curve and change the curve id for each applied voltage. */
	 
	 /* Write the plot details on the first iteration */
	 if(step == 0) {
	   /* Electrostatic Potential */
	   lib->put("output.curve(phi0).xaxis.label", "z-axis");
	   lib->put("output.curve(phi0).xaxis.units", "A");
	   lib->put("output.curve(phi0).yaxis.label", "Potential");
	   lib->put("output.curve(phi0).yaxis.units", "V");
	   
	   /* Cation Concentration */
	   lib->put("output.curve(ccnc0).xaxis.label", "z-axis");
	   lib->put("output.curve(ccnc0).xaxis.units", "A");
	   lib->put("output.curve(ccnc0).yaxis.label", "ion concentration");
	   lib->put("output.curve(ccnc0).yaxis.units", "M");
	 
	   /* Anion Concentration */
	   lib->put("output.curve(acnc0).xaxis.label", "z-axis");
	   lib->put("output.curve(acnc0).xaxis.units", "A");
	   lib->put("output.curve(acnc0).yaxis.label", "ion concentration");
	   lib->put("output.curve(acnc0).yaxis.units", "M");
	 }
	 
	 /* Electrostatic Potential */
	 phipath << "output.curve(phi" << step << ").about.group";
	 lib->put(phipath.str(), "Electrostatic Potential");
	 phipath.str("");
	 
	 voltg_label << "Potential @" << voltg << "V applied";
	 phipath << "output.curve(phi" << step << ").about.label";
	 lib->put(phipath.str(), voltg_label.str());
	 phipath.str("");
	 voltg_label.str("");
	 
	 /* Cation conc. */
	 catpath << "output.curve(ccnc" << step << ").about.group";
	 lib->put(catpath.str(), "Ion Concentrations");
	 catpath.str("");
	 
	 voltg_label << "Cation Concentration @" << voltg << "V applied";
	 catpath << "output.curve(ccnc" << step << ").about.label";
	 lib->put(catpath.str(), voltg_label.str());
	 catpath.str("");
	 voltg_label.str("");
	 
	 /* Anion conc */
	 anipath << "output.curve(acnc" << step << ").about.group";
	 lib->put(anipath.str(), "Ion Concentrations");
	 anipath.str("");
	 
	 voltg_label << "Anion Concentration @" << voltg << "V applied";
	 anipath << "output.curve(acnc" << step << ").about.label";
	 lib->put(anipath.str(), voltg_label.str());
	 anipath.str("");
	 voltg_label.str("");
	 
	 /* set curve id for writing points */
	 phipath << "output.curve(phi" << step << ").component.xy";
	 catpath << "output.curve(ccnc" << step << ").component.xy";
	 anipath << "output.curve(acnc" << step << ").component.xy";
	 
	 cout << "************************************************************" << endl
	      << "*                                                          *" << endl
	      << "*           Writing Potential and Concentrations           *" << endl
	      << "*                                                          *" << endl
	      << "************************************************************" << endl;
	 
	 /* write the concentrations and potential along z-axis */
	 for(int k = 0; k < Ngrdz; k++) {
	   int i = Ngrdx_2; /* x = 0 */
	   int j = Ngrdy_2; /* y = 0 */
	   int iidp = i*Ngrdx + j;
	   int kk = k + (step*Ngrdz);		    
	   z = (k - Ngrdz_2) * Grdlng;
	   
	   phi_val << z << " " << phi_master[iidp][kk] << endl;
	   cat_val << z << " " << ccnc_avg_master[kk] << endl;
	   ani_val << z << " " << acnc_avg_master[kk] << endl;
	 }
	 lib->put(phipath.str(), phi_val.str(), "", 1);
	 lib->put(catpath.str(), cat_val.str(), "", 1);
	 lib->put(anipath.str(), ani_val.str(), "", 1);
	 /* reset datapoint and path strings*/
	 phi_val.str("");
	 cat_val.str("");
	 ani_val.str("");
	 phipath.str("");
	 catpath.str("");
	 anipath.str("");
	 
	 /* if any profiles are being made, find the max value in the array */
	 if(zy_profile || xy_profile || xyz_profile) {
	   cmax = amax = pmin = 0.0;
	   for(int ij = 0; ij < NgrdxNgrdy; ij++) for(int k = 0; k < Ngrdz; k++) {
	     int kk = k + (step*Ngrdz);
	     /* the potential is negative, so we want the min not the max */
	     if(pmin < -phi_master[ij][kk]) {
	       pmin = -phi_master[ij][kk];
	     }
	     if(cmax < ccnc_master[ij][kk]) {
	       cmax = ccnc_master[ij][kk];
	     }
	     if(amax < acnc_master[ij][kk]) {
	       amax = acnc_master[ij][kk];
	     }
	   }
	 }
	 
	 if(zy_profile) {
	   
	   /* Cation conc */
	   /* write label */
	   voltg_label << "Cation Concentration in zy-Plane @ x= " << xcut << "A and " << voltg << "V applied";
	   catpath << "output.field(ccnc_zy" << step << ").about.label";
	   lib->put(catpath.str(), voltg_label.str());
	   catpath.str("");
	   /* set mesh */
	   catpath << "output.field(ccnc_zy" << step << ").component.mesh";
	   lib->put(catpath.str(), "output.cloud(zy-plane)");
	   catpath.str("");
	   voltg_label.str("");
	   
	   /* Anion conc */
	   /* write label */
	   voltg_label << "Anion Concentration in zy-Plane @ x= " << xcut << "A and " << voltg << "V applied";
	   anipath << "output.field(acnc_zy" << step << ").about.label";
	   lib->put(anipath.str(), voltg_label.str());
	   anipath.str("");
	   /* set mesh */
	   anipath << "output.field(acnc_zy" << step << ").component.mesh";
	   lib->put(anipath.str(), "output.cloud(zy-plane)");
	   anipath.str("");
	   voltg_label.str("");
	   
	   /* Potential */
	   /* write label */
	   voltg_label << "Potential Profile in zy-Plane @ x= " << xcut << "A and " << voltg << "V applied";
	   phipath << "output.field(phi_zy" << step << ").about.label";
	   lib->put(phipath.str(), voltg_label.str());
	   phipath.str("");
	   /* set mesh */
	   phipath << "output.field(phi_zy" << step << ").component.mesh";
	   lib->put(phipath.str(), "output.cloud(zy-plane)");
	   phipath.str("");
	   voltg_label.str("");
	   
	   cout << "************************************************************" << endl
		<< "*                                                          *" << endl
		<< "*         Writing Concentration Profile in zy-Plane        *" << endl
		<< "*                                                          *" << endl
		<< "************************************************************" << endl;
	   
	   catpath << "output.field(ccnc_zy" << step << ").component.values";
	   anipath << "output.field(acnc_zy" << step << ").component.values";
	   phipath << "output.field(phi_zy" << step << ").component.values";
	   
	   /* Write Concentration Profile in zy-plane */
	   for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
	     int i = int((xcut/Grdlng)+Ngrdx_2);
	     int iidp = i*Ngrdy + j;
	     idp = i*NgrdyNgrdz + j*Ngrdz + k;
	     int kk = k + (step*Ngrdz);
	     y = (j - Ngrdy_2) * Grdlng;
	     z = (k - Ngrdz_2) * Grdlng;
	     
	     datapoint << y << " " << z << endl;
	     phi_val << (-phi_master[iidp][kk]/pmin) << endl;
	     
	     /* put cnc values in water grids */
	     if(Grdid[idp] == 0 || Grdid[idp] == 4) {
	       cat_val << (ccnc_master[iidp][kk]/cmax) << endl;
	       ani_val << (acnc_master[iidp][kk]/amax) << endl;
	     }
	     /* set color in membrane and ion exclusion region */
	     else if(Grdid[idp] == 3 || Grdid[idp] == 2) {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     /* set color in protein */
	     else {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     /* buffer data */
	     if(idp % 1000 == 0) {
          if(step == 0)
	          lib->put("output.cloud(zy-plane).points", datapoint.str(), "", 1);
	       lib->put(catpath.str(), cat_val.str(), "", 1);
	       lib->put(anipath.str(), ani_val.str(), "", 1);
	       lib->put(phipath.str(), phi_val.str(), "", 1);
	       /* clear buffers */
	       datapoint.str("");
	       cat_val.str("");
	       ani_val.str("");
	       phi_val.str("");
	     }
	   }
	   /* empty remainder of data */
      if(step == 0)
	      lib->put("output.cloud(zy-plane).points", datapoint.str(), "", 1);
	   lib->put(catpath.str(), cat_val.str(), "", 1);
	   lib->put(anipath.str(), ani_val.str(), "", 1);
	   lib->put(phipath.str(), phi_val.str(), "", 1);
	   /* clear buffers */
	   datapoint.str("");
	   cat_val.str("");
	   ani_val.str("");
	   phi_val.str("");
	   catpath.str("");
	   anipath.str("");
	   phipath.str("");
	 }
	 
	 if(xy_profile) {     
	   
	   /* Cation conc */
	   /* write label */
	   voltg_label << "Cation Concentration in xy-plane @ z= " << zcut << "A and " << voltg << "V applied";
	   catpath << "output.field(ccnc_xy" << step << ").about.label";
	   lib->put(catpath.str(), voltg_label.str());
	   catpath.str("");
	   /* set mesh */
	   catpath << "output.field(ccnc_xy" << step << ").component.mesh";
	   lib->put(catpath.str(), "output.cloud(xy-plane)");
	   catpath.str("");
	   voltg_label.str("");
	   
	   /* Anion conc */
	   /* write label */
	   voltg_label << "Anion Concentration in xy-plane @ z= " << zcut << "A and " << voltg << "V applied";
	   anipath << "output.field(acnc_xy" << step << ").about.label";
	   lib->put(anipath.str(), voltg_label.str());
	   anipath.str("");
	   /* set mesh */
	   anipath << "output.field(acnc_xy" << step << ").component.mesh";
	   lib->put(anipath.str(), "output.cloud(xy-plane)");
	   anipath.str("");
	   voltg_label.str("");
	   
	   /* Potential */
	   /* write label */
	   voltg_label << "Potential Profile in xy-Plane @ z= " << zcut << "A and " << voltg << "V applied";
	   phipath << "output.field(phi_xy" << step << ").about.label";
	   lib->put(phipath.str(), voltg_label.str());
	   phipath.str("");
	   /* set mesh */
	   phipath << "output.field(phi_xy" << step << ").component.mesh";
	   lib->put(phipath.str(), "output.cloud(xy-plane)");
	   phipath.str("");
	   voltg_label.str("");
	   
	   cout << "************************************************************" << endl
		<< "*                                                          *" << endl
		<< "*         Writing Concentration Profile in xy-Plane        *" << endl
		<< "*                                                          *" << endl
		<< "************************************************************" << endl;

	   catpath << "output.field(ccnc_xy" << step <<").component.values";
	   anipath << "output.field(acnc_xy" << step <<").component.values";
	   phipath << "output.field(phi_xy"  << step <<").component.values";
	   
	   /* Write Concentration Profile in xy-plane */
	   for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) {
	     int k = int((zcut/Grdlng)+Ngrdz_2);
	     int iidp = i*Ngrdy + j;
	     idp = i*NgrdyNgrdz + j*Ngrdz + k;
	     int kk = k + (step*Ngrdz);
	     y = (j - Ngrdy_2) * Grdlng;		    
	     x = (i - Ngrdx_2) * Grdlng;
	     
	     datapoint << x << " " << y << endl;
	     phi_val << (-phi_master[iidp][kk]/pmin) << endl;
	     
	     /* put cnc values in water grids */
	     if(Grdid[idp] == 0 || Grdid[idp] == 4) {
	       cat_val << (ccnc_master[iidp][kk]/cmax) << endl;
	       ani_val << (acnc_master[iidp][kk]/amax) << endl;
	     }
	     /* set color in membrane and ion exclusion region */
	     else if(Grdid[idp] == 3 || Grdid[idp] == 2) {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     /* set color in protein */
	     else {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     /* buffer data */
	     if(idp % 1000 == 0) {
          if(step == 0)
	          lib->put("output.cloud(xy-plane).points", datapoint.str(), "", 1);
	       lib->put(catpath.str(), cat_val.str(), "", 1);
	       lib->put(anipath.str(), ani_val.str(), "", 1);
	       lib->put(phipath.str(), phi_val.str(), "", 1);
	       /* clear buffers */
	       datapoint.str("");
	       cat_val.str("");
	       ani_val.str("");
	       phi_val.str("");
	     }
	   }
	   /* empty remainder of data */
      if(step == 0)
	      lib->put("output.cloud(xy-plane).points", datapoint.str(), "", 1);
	   lib->put(catpath.str(), cat_val.str(), "", 1);
	   lib->put(anipath.str(), ani_val.str(), "", 1);
	   lib->put(phipath.str(), phi_val.str(), "", 1);
	   /* clear buffers */
	   datapoint.str("");
	   cat_val.str("");
	   ani_val.str("");
	   phi_val.str("");
	   catpath.str("");
	   anipath.str("");
	   phipath.str("");
	 }
	 
	 if(xyz_profile) {
	   
	   /* Cation conc */
	   /* write label */
	   voltg_label << "3D Cation Concentration @" << voltg << "V applied";
	   catpath << "output.field(ccnc_xyz" << step << ").about.label";
	   lib->put(catpath.str(), voltg_label.str());
	   catpath.str("");
	   /* set mesh */
	   catpath << "output.field(ccnc_xyz" << step << ").component.mesh";
	   lib->put(catpath.str(), "output.cloud(xyz)");
	   catpath.str("");
	   voltg_label.str("");
	   
	   /* Anion conc */
	   /* write label */
	   voltg_label << "3D Anion Concentration @" << voltg << "V applied";
	   anipath << "output.field(acnc_xyz" << step << ").about.label";
	   lib->put(anipath.str(), voltg_label.str());
	   anipath.str("");
	   /* set mesh */
	   anipath << "output.field(acnc_xyz" << step << ").component.mesh";
	   lib->put(anipath.str(), "output.cloud(xyz)");
	   anipath.str("");
	   voltg_label.str("");
	   
	   /* Potential */
	   /* write label */
	   voltg_label << "3D Potential @" << voltg << "V applied";
	   phipath << "output.field(phi_xyz" << step << ").about.label";
	   lib->put(phipath.str(), voltg_label.str());
	   phipath.str("");
	   /* set mesh */
	   phipath << "output.field(phi_xyz" << step << ").component.mesh";
	   lib->put(phipath.str(), "output.cloud(xyz)");
	   phipath.str("");
	   voltg_label.str("");
	   
	   cout << "************************************************************" << endl
		<< "*                                                          *" << endl
		<< "*                    Writing 3D Profile                    *" << endl
		<< "*                                                          *" << endl
		<< "************************************************************" << endl;
	   
	   catpath << "output.field(ccnc_xyz" << step << ").component.values";
	   anipath << "output.field(acnc_xyz" << step << ").component.values";
	   phipath << "output.field(phi_xyz" << step << ").component.values";       
	   
	   /* Write 3D Profile */
	   for(int i = 0; i < Ngrdx; i++) for(int j = 0; j < Ngrdy; j++) for(int k = 0; k < Ngrdz; k++) {
	     int iidp = i*Ngrdy + j;
	     idp = i*NgrdyNgrdz + j*Ngrdz + k;
	     int kk = k + (step*Ngrdz);
	     x = (i - Ngrdx_2) * Grdlng;
	     y = (j - Ngrdy_2) * Grdlng;
	     z = (k - Ngrdz_2) * Grdlng;
	     
	     datapoint << x << " " << y << " " << z << endl;
	     phi_val << (-phi_master[iidp][kk]/pmin) << endl;	 
	     
	     /* put cnc values in water grids */
	     if(Grdid[idp] == 0 || Grdid[idp] == 4) {
	       cat_val << (ccnc_master[iidp][kk]/cmax) << endl;
	       ani_val << (acnc_master[iidp][kk]/amax) << endl;
	     }
	     /* set color in membrane and ion exclusion region */
	     else if(Grdid[idp] == 3 || Grdid[idp] == 2) {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     /* set color in protein */
	     else {
	       cat_val << 1.0 << endl;
	       ani_val << 1.0 << endl;
	     }
	     
	     /* buffer data */
	     if(idp % 10000 == 0) {
          if(step == 0)
	          lib->put("output.cloud(xyz).points", datapoint.str(), "", 1);	 
	       lib->put(catpath.str(), cat_val.str(), "", 1);
	       lib->put(anipath.str(), ani_val.str(), "", 1);
	       lib->put(phipath.str(), phi_val.str(), "", 1);
	       /* clear buffers */
	       datapoint.str("");
	       cat_val.str("");
	       ani_val.str("");
	       phi_val.str("");
	     }
	   }
	   /* empty remainder of data */
      if(step == 0)
	      lib->put("output.cloud(xyz).points", datapoint.str(), "", 1);	 
	   lib->put(catpath.str(), cat_val.str(), "", 1);
	   lib->put(anipath.str(), ani_val.str(), "", 1);
	   lib->put(phipath.str(), phi_val.str(), "", 1);
	   /* clear buffers */
	   datapoint.str("");
	   cat_val.str("");
	   ani_val.str("");
	   phi_val.str("");
	   catpath.str("");
	   anipath.str("");
	   phipath.str("");
	 }
       }
       /* End of voltage iteration */   
       /* Send Results and Signal GUI */  
       cout << "************************************************************" << endl
	    << "*                                                          *" << endl
	    << "*     sending results to library, creating run file...     *" << endl
	    << "*                                                          *" << endl
	    << "************************************************************" << endl;
     }
     if(Calc_unstable){
       error << "One or more of the calculations being performed became unstable.  This is a difficult to predict but rare" << endl
	     << "consequence of numerically solving equations using the finite difference method.  It probably means that" << endl 
	     << "a high tolerance setting was paired with a low-denisty grid (Ex. 1e-6 tolerance with the 'coarse' grid)." << endl
	     << "Try changing the numerical parameters and then re-try the simulation." << endl; 
       lib->put("output.string(error).about.label","Error in Calculation"); 
       lib->put("output.string(error).current",error.str());
     }
    lib->result();
   }
   
   /****************************************************************\
    *                                                                *
    *                    Free memory allocations                     *
    *                                                                *
    \****************************************************************/
   free_memory();   
   
   delete[] ccnc_avg;
   delete[] acnc_avg;
   delete[] cflx_avg;
   delete[] aflx_avg;
   delete[] ccrt_avg;
   delete[] acrt_avg;
   delete[] ccnc_avg_master;
   delete[] acnc_avg_master;
   delete[] cflx_avg_master;
   delete[] aflx_avg_master;
   delete[] ccrt_avg_master;
   delete[] acrt_avg_master;
   for(int ij = 0; ij < NgrdxNgrdy; ij++) {
     delete[] phi_master[ij];
     delete[] ccnc_master[ij];
     delete[] acnc_master[ij];
   }
   delete[] phi_master;
   delete[] ccnc_master;
   delete[] acnc_master;

   MPI_Finalize();
   return 0;
}
